@@ -230,7 +230,6 @@ path_equal (const char *a, const char *b, gboolean relax)
 {
   gboolean res;
   size_t a_len, b_len;
-  char *ua, *ub;
 
   if (relax == FALSE)
     return str_equal (a, b, FALSE);
@@ -238,26 +237,20 @@ path_equal (const char *a, const char *b, gboolean relax)
   if (a == NULL || b == NULL)
       return a == b;
 
-  ua = g_uri_unescape_string (a, "/");
-  ub = g_uri_unescape_string (b, "/");
+  a_len = strlen (a);
+  b_len = strlen (b);
 
-  a_len = strlen (ua);
-  b_len = strlen (ub);
-
-  while (a_len > 0 && ua[a_len - 1] == '/')
+  while (a_len > 0 && a[a_len - 1] == '/')
     a_len--;
 
-  while (b_len > 0 && ub[b_len - 1] == '/')
+  while (b_len > 0 && b[b_len - 1] == '/')
     b_len--;
 
   if (a_len == b_len)
-    res = ! strncmp (ua, ub, a_len);
+    res = ! strncmp (a, b, a_len);
   else
     res = FALSE;
 
-  g_free(ua);
-  g_free(ub);
-
   return res;
 }
 
@@ -266,13 +259,20 @@ static gboolean
 dav_uri_match (SoupURI *a, SoupURI *b, gboolean relax)
 {
   gboolean diff;
+  char *ua, *ub;
+
+  ua = g_uri_unescape_string (a->path, "/");
+  ub = g_uri_unescape_string (b->path, "/");
 
   diff = a->scheme != b->scheme || a->port != b->port  ||
     ! str_equal (a->host, b->host, TRUE)               ||
-    ! path_equal (a->path, b->path, relax)             ||
+    ! path_equal (ua, ub, relax)                       ||
     ! str_equal (a->query, b->query, FALSE)            ||
     ! str_equal (a->fragment, b->fragment, FALSE);
 
+  g_free (ua);
+  g_free (ub);
+
   return !diff;
 }
 
@@ -1387,10 +1387,12 @@ stat_location (GVfsBackend  *backend,
         	           http_error_code_from_status (status),
                 	   msg->reason_phrase);
 
+      g_object_unref (msg);
       return FALSE;
     }
 
   res = stat_location_finish (msg, target_type, num_children);
+  g_object_unref (msg);
 
   if (res == FALSE)
     g_set_error_literal (error, 
@@ -2620,6 +2622,7 @@ do_delete (GVfsBackend   *backend,
     {
       g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
       g_error_free (error);
+      soup_uri_free (uri);
       return;
     }
 
@@ -2628,6 +2631,7 @@ do_delete (GVfsBackend   *backend,
       g_vfs_job_failed (G_VFS_JOB (job),
                         G_IO_ERROR, G_IO_ERROR_NOT_EMPTY,
                         _("Directory not empty"));
+      soup_uri_free (uri);
       return;
     }
 
@@ -2753,8 +2757,7 @@ do_move (GVfsBackend *backend,
   if (!res && !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
     {
       g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
-      g_error_free (error);
-      return;
+      goto error;
     }
 
   if (res)
@@ -2776,7 +2779,7 @@ do_move (GVfsBackend *backend,
                                               G_IO_ERROR,
                                               G_IO_ERROR_IS_DIRECTORY,
                                               _("Can't move over directory"));
-                  return;
+                  goto error;
                 }
               else if (source_ft == G_FILE_TYPE_DIRECTORY)
                 {
@@ -2792,7 +2795,7 @@ do_move (GVfsBackend *backend,
                     {
                       http_job_failed (G_VFS_JOB (job), msg);
                       g_object_unref (msg);
-                      return;
+                      goto error;
                     }
                   g_object_unref (msg);
                 }
@@ -2800,8 +2803,7 @@ do_move (GVfsBackend *backend,
           else
             {
               g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
-              g_error_free (error);
-              return;
+              goto error;
             }
         }
       else
@@ -2810,6 +2812,7 @@ do_move (GVfsBackend *backend,
                                     G_IO_ERROR,
                                     G_IO_ERROR_EXISTS,
                                     _("Target file exists"));
+          goto error;
         }
     }
 
@@ -2818,21 +2821,8 @@ do_move (GVfsBackend *backend,
 
   status = g_vfs_backend_dav_send_message (backend, msg);
 
-  /*
-   * The precondition of SOUP_STATUS_PRECONDITION_FAILED (412) in
-   * this case was triggered by the "Overwrite: F" header which
-   * means that the target already exists.
-   * Also if we get a REDIRECTION it means that there was no
-   * "Location" header, since otherwise that would have triggered
-   * our redirection handler. This probably means we are dealing
-   * with an web dav implementation (like mod_dav) that also sends
-   * redirects for the destionaion (i.e. "Destination: /foo" header)
-   * which very likely means that the target also exists (and is a
-   * directory). That or the webdav server is broken.
-   * We could find out by doing another stat and but I think this is
-   * such a corner case that we are totally fine with returning
-   * G_IO_ERROR_EXISTS.
-   * */
+  /* See do_set_display_name () for the explanation of the PRECONDITION_FAILED
+   * and IS_REDIRECTION handling below. */
 
   if (SOUP_STATUS_IS_SUCCESSFUL (status))
     {
@@ -2846,7 +2836,119 @@ do_move (GVfsBackend *backend,
   else
     http_job_failed (G_VFS_JOB (job), msg);
 
+error:
   g_object_unref (msg);
+  g_clear_error (&error);
+  soup_uri_free (source_uri);
+  soup_uri_free (target_uri);
+}
+
+static void
+do_copy (GVfsBackend *backend,
+         GVfsJobCopy *job,
+         const char *source,
+         const char *destination,
+         GFileCopyFlags flags,
+         GFileProgressCallback progress_callback,
+         gpointer progress_callback_data)
+{
+  SoupMessage *msg;
+  SoupURI *source_uri;
+  SoupURI *target_uri;
+  guint status;
+  GFileType source_ft, target_ft;
+  GError *error = NULL;
+  gboolean res;
+
+  if (flags & G_FILE_COPY_BACKUP)
+    {
+      /* Return G_IO_ERROR_NOT_SUPPORTED instead of
+       * G_IO_ERROR_CANT_CREATE_BACKUP to proceed with the GIO fallback
+       * copy. */
+      g_vfs_job_failed_literal (G_VFS_JOB (job),
+                                G_IO_ERROR,
+                                G_IO_ERROR_NOT_SUPPORTED,
+                                "Operation not supported");
+      return;
+    }
+
+  source_uri = g_vfs_backend_dav_uri_for_path (backend, source, FALSE);
+  target_uri = g_vfs_backend_dav_uri_for_path (backend, destination, FALSE);
+
+  res = stat_location (backend, source_uri, &source_ft, NULL, &error);
+  if (!res)
+    {
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+      goto error;
+    }
+
+  res = stat_location (backend, target_uri, &target_ft, NULL, &error);
+  if (res)
+    {
+      if (flags & G_FILE_COPY_OVERWRITE)
+        {
+          if (target_ft == G_FILE_TYPE_DIRECTORY)
+            {
+              if (source_ft == G_FILE_TYPE_DIRECTORY)
+                g_vfs_job_failed_literal (G_VFS_JOB(job),
+                                          G_IO_ERROR,
+                                          G_IO_ERROR_WOULD_MERGE,
+                                          _("Can't copy directory over directory"));
+              else
+                g_vfs_job_failed_literal (G_VFS_JOB(job),
+                                          G_IO_ERROR,
+                                          G_IO_ERROR_IS_DIRECTORY,
+                                          _("File is directory"));
+              goto error;
+            }
+        }
+      else
+        {
+          g_vfs_job_failed_literal (G_VFS_JOB (job),
+                                    G_IO_ERROR,
+                                    G_IO_ERROR_EXISTS,
+                                    _("Target file already exists"));
+          goto error;
+        }
+    }
+  else if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
+    {
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+      goto error;
+    }
+
+  if (source_ft == G_FILE_TYPE_DIRECTORY)
+    {
+      g_vfs_job_failed_literal (G_VFS_JOB (job),
+                                G_IO_ERROR,
+                                G_IO_ERROR_WOULD_RECURSE,
+                                _("Can't recursively copy directory"));
+      goto error;
+    }
+
+  msg = soup_message_new_from_uri (SOUP_METHOD_COPY, source_uri);
+  message_add_destination_header (msg, target_uri);
+  message_add_overwrite_header (msg, flags & G_FILE_COPY_OVERWRITE);
+
+  status = g_vfs_backend_dav_send_message (backend, msg);
+
+  /* See do_set_display_name () for the explanation of the PRECONDITION_FAILED
+   * and IS_REDIRECTION handling below. */
+
+  if (SOUP_STATUS_IS_SUCCESSFUL (status))
+    g_vfs_job_succeeded (G_VFS_JOB (job));
+  else if (status == SOUP_STATUS_PRECONDITION_FAILED ||
+           SOUP_STATUS_IS_REDIRECTION (status))
+    g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                      G_IO_ERROR_EXISTS,
+                      _("Target file already exists"));
+  else
+    http_job_failed (G_VFS_JOB (job), msg);
+
+  g_object_unref (msg);
+
+error:
+  g_clear_error (&error);
   soup_uri_free (source_uri);
   soup_uri_free (target_uri);
 }
@@ -3255,5 +3357,6 @@ g_vfs_backend_dav_class_init (GVfsBackendDavClass *klass)
   backend_class->delete            = do_delete;
   backend_class->set_display_name  = do_set_display_name;
   backend_class->move              = do_move;
+  backend_class->copy              = do_copy;
   backend_class->try_push          = try_push;
 }
